/* Front-end language hooks for the quine language.
   Copyright (c) 2013 Free Software Foundation, Inc.
   Copyright (c) 2014 Christopher Smith

This file is part of gquine.

gquine is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

gquine is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
gquine. If not, see <http://www.gnu.org/licenses/>. */

/* Basic stuff. */
#include "config.h"
#include "system.h"
#include "ansidecl.h"
/* Trees. */
#include "coretypes.h"
#include "tree.h"
/* Stuff we do with trees. */
#include "tree-iterator.h"
#include "dumpfile.h"
#include "gimple.h"
#include "cgraph.h"
/* Options. */
#include "opts.h"
#include "options.h"
#include "flags.h"
/* Warnings and errors. */
#include "diagnostic-core.h"
/* Hooks. */
#include "toplev.h"
#include "langhooks.h"
#include "langhooks-def.h"
/* GGC needs this. */
#include "debug.h"
/* Vector template class. */
#include "vec.h"

/* Required structures. */

struct GTY(()) lang_decl
{
  char dummy;
};

struct GTY(()) lang_type
{
  char dummy;
};

struct GTY(()) language_function
{
  char dummy;
};

struct GTY(()) lang_identifier
{
  struct tree_identifier common;
};

union GTY((desc ("0"),
           chain_next ("(union lang_tree_node *) TREE_CHAIN (&%h.generic)")))
      lang_tree_node
{
  union tree_node GTY((tag ("0"), desc ("tree_node_structure (&%h)"))) generic;
};

/* Global variables. */

tree global_decls = NULL_TREE;

/* Language hooks. */

#undef LANG_HOOKS_NAME
#define LANG_HOOKS_NAME "GNU Quine"

#undef LANG_HOOKS_OPTION_LANG_MASK
#define LANG_HOOKS_OPTION_LANG_MASK quine_option_lang_mask
unsigned int
quine_option_lang_mask (void)
{
  return CL_quine;
}

#undef LANG_HOOKS_INIT
#define LANG_HOOKS_INIT quine_init
bool
quine_init (void)
{
  /* Create default types like int. */
  build_common_tree_nodes (false, false);
  return true;
}

/* Language hook to read and parse the source file. */
#undef LANG_HOOKS_PARSE_FILE
#define LANG_HOOKS_PARSE_FILE quine_parse_file
void
quine_parse_file (void)
{
  /* Open the source file. */
  FILE *in = fopen (in_fnames[0], "r");
  if (in == NULL)
    {
      error ("Can't open input file %s", in_fnames[0]);
      return;
    }

  /* Read the file into an array. */
  vec<char> file_data = vNULL;
  int c;
  while ((c = fgetc (in)) != EOF)
    file_data.safe_push (c);
  /* If the file ends in a newline, remove it; puts () will append a newline
   * anyway. If the file does not end in a newline, puts () will add one, but
   * this is a Good Thing when run in a terminal. */
  if (!file_data.is_empty () && file_data.last () == '\n')
    file_data.pop ();
  file_data.safe_push (0);

  fclose (in);

  /* Create the main () function. */
  /* Type for a function which returns int and has no parameters. */
  tree main_type = build_function_type_list (integer_type_node, NULL_TREE);
  /* Function with no source location, named "main" and of the above type. */
  tree main_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                               get_identifier ("main"), main_type);

  /* File scope. */
  DECL_CONTEXT (main_decl) = NULL_TREE;
  /* Has a definition. */
  TREE_STATIC (main_decl) = true;
  /* External linkage. */
  TREE_PUBLIC (main_decl) = true;
  /* Don't need anywhere to store argumnets. */
  DECL_ARGUMENTS (main_decl) = NULL_TREE;

  /* Return value for main (). */
  /* Result variable with no name and same type as main () returns. */
  tree main_ret = build_decl (BUILTINS_LOCATION, RESULT_DECL, NULL_TREE,
                              TREE_TYPE (main_type));
  /* Scoped within the main function. */
  DECL_CONTEXT (main_ret) = main_decl;
  /* Generated by the compiler. */
  DECL_ARTIFICIAL (main_ret) = true;
  /* No debugging symbol. */
  DECL_IGNORED_P (main_ret) = true;

  /* The result of the function is the above value. */
  DECL_RESULT (main_decl) = main_ret;

  /* Block to represent the scope of local variables. */
  tree bl = build_block (NULL_TREE, NULL_TREE, main_decl, NULL_TREE);
  DECL_INITIAL (main_decl) = bl;
  TREE_USED (bl) = true;

  /* The bind expression contains the statements to execute. */
  tree bind = build3 (BIND_EXPR, void_type_node, BLOCK_VARS (bl), NULL_TREE,
                      bl);
  /* Don't optimise it away. */
  TREE_SIDE_EFFECTS (bind) = true;

  /* List of statements in the main () function. */
  tree main_stmts = alloc_stmt_list ();

  /* Build the puts () function declaration. */
  /* Function type which returns int with unspecified parameters. */
  tree puts_type = build_function_type (integer_type_node, NULL_TREE);
  /* Function named "puts" of that type. */
  tree puts_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                               get_identifier ("puts"), puts_type);
  /* Only a declaration, no definition. */
  DECL_EXTERNAL (puts_decl) = true;
  /* External linkage. */
  TREE_PUBLIC (puts_decl) = true;

  /* puts("..."); */
  /* Expression which calls puts () with 1 argument which is a string literal
   * containing the source file data. */
  tree call_puts = build_call_expr (puts_decl, 1,
                                    build_string_literal (
                                        file_data.length (),
                                        file_data.address ()));
  append_to_statement_list (call_puts, &main_stmts);

  /* return 0; */
  /* Assign 0 to the return value. */
  tree main_set_ret = build2 (MODIFY_EXPR, TREE_TYPE (main_ret), main_ret,
                              build_int_cst (integer_type_node, 0));
  /* Perform the assignment and return from the function. */
  tree main_ret_expr = build1 (RETURN_EXPR, void_type_node, main_set_ret);
  append_to_statement_list (main_ret_expr, &main_stmts);

  /* Make the bind contain the statements. */
  BIND_EXPR_BODY (bind) = main_stmts;
  /* Set main to use the statements in bind. */
  DECL_SAVED_TREE (main_decl) = bind;

  /* Add the main function to the list of global declarations. */
  global_decls = chainon (main_decl, global_decls);

  /* Dump out the translation unit. */
  FILE *tu_stream = dump_begin (TDI_tu, NULL);
  if (tu_stream)
    {
      dump_node (global_decls, 0, tu_stream);
      dump_end (TDI_tu, tu_stream);
    }

  /* Prepare declarations for middle-end. */
  for (tree t = global_decls; t; t = TREE_CHAIN (t))
    {
      gimplify_function_tree (t);
      cgraph_finalize_function (t, false);
    }
}

/* Unused; terminate with ICE. */
#undef LANG_HOOKS_GLOBAL_BINDINGS_P
#define LANG_HOOKS_GLOBAL_BINDINGS_P quine_global_bindings_p
bool
quine_global_bindings_p (void)
{
  gcc_unreachable ();
}

/* Unused; terminate with ICE. */
#undef LANG_HOOKS_PUSHDECL
#define LANG_HOOKS_PUSHDECL quine_pushdecl
tree
quine_pushdecl (tree decl ATTRIBUTE_UNUSED)
{
  gcc_unreachable ();
}

/* Return the global declarations. */
#undef LANG_HOOKS_GETDECLS
#define LANG_HOOKS_GETDECLS quine_getdecls
tree
quine_getdecls (void)
{
  return global_decls;
}

/* Unused; terminate with ICE. */
#undef LANG_HOOKS_TYPE_FOR_MODE
#define LANG_HOOKS_TYPE_FOR_MODE quine_type_for_mode
tree
quine_type_for_mode (enum machine_mode mode ATTRIBUTE_UNUSED,
                     int unsignedp ATTRIBUTE_UNUSED)
{
  gcc_unreachable ();
}

/* Unused; terminate with ICE. */
#undef LANG_HOOKS_TYPE_FOR_SIZE
#define LANG_HOOKS_TYPE_FOR_SIZE quine_type_for_size
tree
quine_type_for_size (unsigned precision ATTRIBUTE_UNUSED,
                     int unsignedp ATTRIBUTE_UNUSED)
{
  gcc_unreachable ();
}

/* Define the lang_hooks structure. */

struct lang_hooks lang_hooks = LANG_HOOKS_INITIALIZER;

/* GGC stuff. */

#include "gt-quine-quine1.h"
#include "gtype-quine.h"
